@@ -30,7 +30,6 @@ from ..mesonlib import (
 from .c_function_attributes import C_FUNC_ATTRIBUTES
 
 from .compilers import (
-    CompilerType,
     get_largefile_args,
     gnu_winlibs,
     msvc_winlibs,
@@ -46,6 +45,7 @@ from .compilers import (
     ElbrusCompiler,
     IntelCompiler,
     RunResult,
+    CcrxCompiler,
 )
 
 gnu_compiler_internal_libs = ('m', 'c', 'pthread', 'dl', 'rt')
@@ -121,10 +121,7 @@ class CCompiler(Compiler):
     # The default behavior is this, override in MSVC
     @functools.lru_cache(maxsize=None)
     def build_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
-        if getattr(self, 'compiler_type', False) and self.compiler_type.is_osx_compiler:
-            # Clang, GCC and ICC on macOS all use the same rpath arguments
-            return self.build_osx_rpath_args(build_dir, rpath_paths, build_rpath)
-        elif self.compiler_type.is_windows_compiler:
+        if self.compiler_type.is_windows_compiler:
             return []
         return self.build_unix_rpath_args(build_dir, from_dir, rpath_paths, build_rpath, install_rpath)
 
@@ -159,23 +156,6 @@ class CCompiler(Compiler):
         '''
         return self.get_no_optimization_args()
 
-    def get_allow_undefined_link_args(self):
-        '''
-        Get args for allowing undefined symbols when linking to a shared library
-        '''
-        if self.id in ('clang', 'gcc'):
-            if self.compiler_type.is_osx_compiler:
-                # Apple ld
-                return ['-Wl,-undefined,dynamic_lookup']
-            else:
-                # GNU ld and LLVM lld
-                return ['-Wl,--allow-shlib-undefined']
-        elif self.id == 'msvc':
-            # link.exe
-            return ['/FORCE:UNRESOLVED']
-        # FIXME: implement other linkers
-        return []
-
     def get_output_args(self, target):
         return ['-o', target]
 
@@ -208,7 +188,9 @@ class CCompiler(Compiler):
     def _get_search_dirs(self, env):
         extra_args = ['--print-search-dirs']
         stdo = None
-        with self._build_wrapper('', env, extra_args, None, 'compile', True) as p:
+        with self._build_wrapper('', env, extra_args=extra_args,
+                                 dependencies=None, mode='compile',
+                                 want_output=True) as p:
             stdo = p.stdo
         return stdo
 
@@ -216,9 +198,25 @@ class CCompiler(Compiler):
     def _split_fetch_real_dirs(pathstr, sep=':'):
         paths = []
         for p in pathstr.split(sep):
-            p = Path(p)
-            if p.exists():
-                paths.append(p.resolve().as_posix())
+            # GCC returns paths like this:
+            # /usr/lib/gcc/x86_64-linux-gnu/8/../../../../x86_64-linux-gnu/lib
+            # It would make sense to normalize them to get rid of the .. parts
+            # Sadly when you are on a merged /usr fs it also kills these:
+            # /lib/x86_64-linux-gnu
+            # since /lib is a symlink to /usr/lib. This would mean
+            # paths under /lib would be considered not a "system path",
+            # which is wrong and breaks things. Store everything, just to be sure.
+            pobj = Path(p)
+            unresolved = pobj.as_posix()
+            if pobj.exists():
+                if unresolved not in paths:
+                    paths.append(unresolved)
+                try:
+                    resolved = Path(p).resolve().as_posix()
+                    if resolved not in paths:
+                        paths.append(resolved)
+                except FileNotFoundError:
+                    pass
         return tuple(paths)
 
     def get_compiler_dirs(self, env, name):
@@ -232,8 +230,34 @@ class CCompiler(Compiler):
         return ()
 
     @functools.lru_cache()
-    def get_library_dirs(self, env):
-        return self.get_compiler_dirs(env, 'libraries')
+    def get_library_dirs(self, env, elf_class = None):
+        dirs = self.get_compiler_dirs(env, 'libraries')
+        if elf_class is None or elf_class == 0:
+            return dirs
+
+        # if we do have an elf class for 32-bit or 64-bit, we want to check that
+        # the directory in question contains libraries of the appropriate class. Since
+        # system directories aren't mixed, we only need to check one file for each
+        # directory and go by that. If we can't check the file for some reason, assume
+        # the compiler knows what it's doing, and accept the directory anyway.
+        retval = []
+        for d in dirs:
+            files = [f for f in os.listdir(d) if f.endswith('.so') and os.path.isfile(os.path.join(d, f))]
+            # if no files, accept directory and move on
+            if len(files) == 0:
+                retval.append(d)
+                continue
+            file_to_check = os.path.join(d, files[0])
+            with open(file_to_check, 'rb') as fd:
+                header = fd.read(5)
+                # if file is not an ELF file, it's weird, but accept dir
+                # if it is elf, and the class matches, accept dir
+                if header[1:4] != b'ELF' or int(header[4]) == elf_class:
+                    retval.append(d)
+                # at this point, it's an ELF file which doesn't match the
+                # appropriate elf_class, so skip this one
+                pass
+        return tuple(retval)
 
     @functools.lru_cache()
     def get_program_dirs(self, env):
@@ -334,13 +358,14 @@ class CCompiler(Compiler):
         code = 'int main(int argc, char **argv) { int class=0; return class; }\n'
         return self.sanity_check_impl(work_dir, environment, 'sanitycheckc.c', code)
 
-    def check_header(self, hname, prefix, env, extra_args=None, dependencies=None):
+    def check_header(self, hname, prefix, env, *, extra_args=None, dependencies=None):
         fargs = {'prefix': prefix, 'header': hname}
         code = '''{prefix}
         #include <{header}>'''
-        return self.compiles(code.format(**fargs), env, extra_args, dependencies)
+        return self.compiles(code.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies)
 
-    def has_header(self, hname, prefix, env, extra_args=None, dependencies=None):
+    def has_header(self, hname, prefix, env, *, extra_args=None, dependencies=None):
         fargs = {'prefix': prefix, 'header': hname}
         code = '''{prefix}
         #ifdef __has_include
@@ -350,10 +375,10 @@ class CCompiler(Compiler):
         #else
          #include <{header}>
         #endif'''
-        return self.compiles(code.format(**fargs), env, extra_args,
-                             dependencies, 'preprocess')
+        return self.compiles(code.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies, mode='preprocess')
 
-    def has_header_symbol(self, hname, symbol, prefix, env, extra_args=None, dependencies=None):
+    def has_header_symbol(self, hname, symbol, prefix, env, *, extra_args=None, dependencies=None):
         fargs = {'prefix': prefix, 'header': hname, 'symbol': symbol}
         t = '''{prefix}
         #include <{header}>
@@ -363,13 +388,16 @@ class CCompiler(Compiler):
                 {symbol};
             #endif
         }}'''
-        return self.compiles(t.format(**fargs), env, extra_args, dependencies)
+        return self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies)
 
     def _get_compiler_check_args(self, env, extra_args, dependencies, mode='compile'):
         if extra_args is None:
             extra_args = []
-        elif isinstance(extra_args, str):
-            extra_args = [extra_args]
+        else:
+            extra_args = listify(extra_args)
+        extra_args = listify([e(mode) if callable(e) else e for e in extra_args])
+
         if dependencies is None:
             dependencies = []
         elif not isinstance(dependencies, list):
@@ -399,7 +427,13 @@ class CCompiler(Compiler):
                 args += env.coredata.get_external_preprocess_args(self.language)
             elif mode == 'compile':
                 # Add CFLAGS/CXXFLAGS/OBJCFLAGS/OBJCXXFLAGS from the env
-                args += env.coredata.get_external_args(self.language)
+                sys_args = env.coredata.get_external_args(self.language)
+                # Apparently it is a thing to inject linker flags both
+                # via CFLAGS _and_ LDFLAGS, even though the former are
+                # also used during linking. These flags can break
+                # argument checks. Thanks, Autotools.
+                cleaned_sys_args = self.remove_linkerlike_args(sys_args)
+                args += cleaned_sys_args
             elif mode == 'link':
                 # Add LDFLAGS from the env
                 args += env.coredata.get_external_link_args(self.language)
@@ -408,7 +442,7 @@ class CCompiler(Compiler):
         args += extra_args
         return args
 
-    def compiles(self, code, env, extra_args=None, dependencies=None, mode='compile'):
+    def compiles(self, code, env, *, extra_args=None, dependencies=None, mode='compile'):
         with self._build_wrapper(code, env, extra_args, dependencies, mode) as p:
             return p.returncode == 0
 
@@ -416,10 +450,11 @@ class CCompiler(Compiler):
         args = self._get_compiler_check_args(env, extra_args, dependencies, mode)
         return self.compile(code, args, mode, want_output=want_output)
 
-    def links(self, code, env, extra_args=None, dependencies=None):
-        return self.compiles(code, env, extra_args, dependencies, mode='link')
+    def links(self, code, env, *, extra_args=None, dependencies=None):
+        return self.compiles(code, env, extra_args=extra_args,
+                             dependencies=dependencies, mode='link')
 
-    def run(self, code, env, extra_args=None, dependencies=None):
+    def run(self, code, env, *, extra_args=None, dependencies=None):
         if self.is_cross and self.exe_wrapper is None:
             raise CrossNoRunException('Can not run test applications in this cross environment.')
         with self._build_wrapper(code, env, extra_args, dependencies, mode='link', want_output=True) as p:
@@ -449,7 +484,8 @@ class CCompiler(Compiler):
         t = '''#include <stdio.h>
         {prefix}
         int main() {{ static int a[1-2*!({expression})]; a[0]=0; return 0; }}'''
-        return self.compiles(t.format(**fargs), env, extra_args, dependencies)
+        return self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies)
 
     def cross_compute_int(self, expression, low, high, guess, prefix, env, extra_args, dependencies):
         # Try user's guess first
@@ -499,7 +535,7 @@ class CCompiler(Compiler):
 
         return low
 
-    def compute_int(self, expression, low, high, guess, prefix, env, extra_args=None, dependencies=None):
+    def compute_int(self, expression, low, high, guess, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         if self.is_cross:
@@ -511,14 +547,15 @@ class CCompiler(Compiler):
             printf("%ld\\n", (long)({expression}));
             return 0;
         }};'''
-        res = self.run(t.format(**fargs), env, extra_args, dependencies)
+        res = self.run(t.format(**fargs), env, extra_args=extra_args,
+                       dependencies=dependencies)
         if not res.compiled:
             return -1
         if res.returncode != 0:
             raise EnvironmentException('Could not run compute_int test binary.')
         return int(res.stdout)
 
-    def cross_sizeof(self, typename, prefix, env, extra_args=None, dependencies=None):
+    def cross_sizeof(self, typename, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         fargs = {'prefix': prefix, 'type': typename}
@@ -527,30 +564,33 @@ class CCompiler(Compiler):
         int main(int argc, char **argv) {{
             {type} something;
         }}'''
-        if not self.compiles(t.format(**fargs), env, extra_args, dependencies):
+        if not self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies):
             return -1
         return self.cross_compute_int('sizeof(%s)' % typename, None, None, None, prefix, env, extra_args, dependencies)
 
-    def sizeof(self, typename, prefix, env, extra_args=None, dependencies=None):
+    def sizeof(self, typename, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         fargs = {'prefix': prefix, 'type': typename}
         if self.is_cross:
-            return self.cross_sizeof(typename, prefix, env, extra_args, dependencies)
+            return self.cross_sizeof(typename, prefix, env, extra_args=extra_args,
+                                     dependencies=dependencies)
         t = '''#include<stdio.h>
         {prefix}
         int main(int argc, char **argv) {{
             printf("%ld\\n", (long)(sizeof({type})));
             return 0;
         }};'''
-        res = self.run(t.format(**fargs), env, extra_args, dependencies)
+        res = self.run(t.format(**fargs), env, extra_args=extra_args,
+                       dependencies=dependencies)
         if not res.compiled:
             return -1
         if res.returncode != 0:
             raise EnvironmentException('Could not run sizeof test binary.')
         return int(res.stdout)
 
-    def cross_alignment(self, typename, prefix, env, extra_args=None, dependencies=None):
+    def cross_alignment(self, typename, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         fargs = {'prefix': prefix, 'type': typename}
@@ -559,7 +599,8 @@ class CCompiler(Compiler):
         int main(int argc, char **argv) {{
             {type} something;
         }}'''
-        if not self.compiles(t.format(**fargs), env, extra_args, dependencies):
+        if not self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies):
             return -1
         t = '''#include <stddef.h>
         {prefix}
@@ -569,11 +610,12 @@ class CCompiler(Compiler):
         }};'''
         return self.cross_compute_int('offsetof(struct tmp, target)', None, None, None, t.format(**fargs), env, extra_args, dependencies)
 
-    def alignment(self, typename, prefix, env, extra_args=None, dependencies=None):
+    def alignment(self, typename, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         if self.is_cross:
-            return self.cross_alignment(typename, prefix, env, extra_args, dependencies)
+            return self.cross_alignment(typename, prefix, env, extra_args=extra_args,
+                                        dependencies=dependencies)
         fargs = {'prefix': prefix, 'type': typename}
         t = '''#include <stdio.h>
         #include <stddef.h>
@@ -586,7 +628,8 @@ class CCompiler(Compiler):
             printf("%d", (int)offsetof(struct tmp, target));
             return 0;
         }}'''
-        res = self.run(t.format(**fargs), env, extra_args, dependencies)
+        res = self.run(t.format(**fargs), env, extra_args=extra_args,
+                       dependencies=dependencies)
         if not res.compiled:
             raise EnvironmentException('Could not compile alignment test.')
         if res.returncode != 0:
@@ -630,7 +673,7 @@ class CCompiler(Compiler):
         int main(int argc, char *argv[]) {{
             printf ("{fmt}", {cast} {f}());
         }}'''.format(**fargs)
-        res = self.run(code, env, extra_args, dependencies)
+        res = self.run(code, env, extra_args=extra_args, dependencies=dependencies)
         if not res.compiled:
             m = 'Could not get return value of {}()'
             raise EnvironmentException(m.format(fname))
@@ -699,7 +742,7 @@ class CCompiler(Compiler):
         }}'''
         return head, main
 
-    def has_function(self, funcname, prefix, env, extra_args=None, dependencies=None):
+    def has_function(self, funcname, prefix, env, *, extra_args=None, dependencies=None):
         """
         First, this function looks for the symbol in the default libraries
         provided by the compiler (stdlib + a few others usually). If that
@@ -747,7 +790,8 @@ class CCompiler(Compiler):
             head, main = self._no_prototype_templ()
         templ = head + stubs_fail + main
 
-        if self.links(templ.format(**fargs), env, extra_args, dependencies):
+        if self.links(templ.format(**fargs), env, extra_args=extra_args,
+                      dependencies=dependencies):
             return True
 
         # MSVC does not have compiler __builtin_-s.
@@ -780,9 +824,10 @@ class CCompiler(Compiler):
             #endif
         #endif
         }}'''
-        return self.links(t.format(**fargs), env, extra_args, dependencies)
+        return self.links(t.format(**fargs), env, extra_args=extra_args,
+                          dependencies=dependencies)
 
-    def has_members(self, typename, membernames, prefix, env, extra_args=None, dependencies=None):
+    def has_members(self, typename, membernames, prefix, env, *, extra_args=None, dependencies=None):
         if extra_args is None:
             extra_args = []
         fargs = {'prefix': prefix, 'type': typename, 'name': 'foo'}
@@ -796,7 +841,8 @@ class CCompiler(Compiler):
             {type} {name};
             {members}
         }};'''
-        return self.compiles(t.format(**fargs), env, extra_args, dependencies)
+        return self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies)
 
     def has_type(self, typename, prefix, env, extra_args, dependencies=None):
         fargs = {'prefix': prefix, 'type': typename}
@@ -804,7 +850,8 @@ class CCompiler(Compiler):
         void bar() {{
             sizeof({type});
         }};'''
-        return self.compiles(t.format(**fargs), env, extra_args, dependencies)
+        return self.compiles(t.format(**fargs), env, extra_args=extra_args,
+                             dependencies=dependencies)
 
     def symbols_have_underscore_prefix(self, env):
         '''
@@ -864,17 +911,17 @@ class CCompiler(Compiler):
         stlibext = ['a']
         # We've always allowed libname to be both `foo` and `libfoo`,
         # and now people depend on it
-        if strict and self.id != 'msvc': # lib prefix is not usually used with msvc
+        if strict and not isinstance(self, VisualStudioCCompiler): # lib prefix is not usually used with msvc
             prefixes = ['lib']
         else:
             prefixes = ['lib', '']
         # Library suffixes and prefixes
         if for_darwin(env.is_cross_build(), env):
-            shlibext = ['dylib']
+            shlibext = ['dylib', 'so']
         elif for_windows(env.is_cross_build(), env):
             # FIXME: .lib files can be import or static so we should read the
             # file, figure out which one it is, and reject the wrong kind.
-            if self.id == 'msvc':
+            if isinstance(self, VisualStudioCCompiler):
                 shlibext = ['lib']
             else:
                 shlibext = ['dll.a', 'lib', 'dll']
@@ -935,6 +982,13 @@ class CCompiler(Compiler):
                 return f
         return None
 
+    @functools.lru_cache()
+    def output_is_64bit(self, env):
+        '''
+        returns true if the output produced is 64-bit, false if 32-bit
+        '''
+        return self.sizeof('void *', '', env) == 8
+
     def find_library_real(self, libname, env, extra_dirs, code, libtype):
         # First try if we can just add the library as -l.
         # Gcc + co seem to prefer builtin lib dirs to -L dirs.
@@ -950,8 +1004,18 @@ class CCompiler(Compiler):
         # Not found or we want to use a specific libtype? Try to find the
         # library file itself.
         patterns = self.get_library_naming(env, libtype)
+        # try to detect if we are 64-bit or 32-bit. If we can't
+        # detect, we will just skip path validity checks done in
+        # get_library_dirs() call
+        try:
+            if self.output_is_64bit(env):
+                elf_class = 2
+            else:
+                elf_class = 1
+        except:
+            elf_class = 0
         # Search in the specified dirs, and then in the system libraries
-        for d in itertools.chain(extra_dirs, self.get_library_dirs(env)):
+        for d in itertools.chain(extra_dirs, self.get_library_dirs(env, elf_class)):
             for p in patterns:
                 trial = self._get_trials_from_pattern(p, d, libname)
                 if not trial:
@@ -986,12 +1050,12 @@ class CCompiler(Compiler):
         return self.find_library_impl(libname, env, extra_dirs, code, libtype)
 
     def thread_flags(self, env):
-        if for_haiku(self.is_cross, env):
+        if for_haiku(self.is_cross, env) or for_darwin(self.is_cross, env):
             return []
         return ['-pthread']
 
     def thread_link_flags(self, env):
-        if for_haiku(self.is_cross, env):
+        if for_haiku(self.is_cross, env) or for_darwin(self.is_cross, env):
             return []
         return ['-pthread']
 
@@ -1094,9 +1158,9 @@ class ClangCCompiler(ClangCompiler, CCompiler):
 
 
 class ArmclangCCompiler(ArmclangCompiler, CCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        ArmclangCompiler.__init__(self)
+        ArmclangCompiler.__init__(self, compiler_type)
         default_warn_args = ['-Wall', '-Winvalid-pch']
         self.warn_args = {'1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
@@ -1136,7 +1200,7 @@ class GnuCCompiler(GnuCompiler, CCompiler):
                                                        ['none', 'c89', 'c99', 'c11',
                                                         'gnu89', 'gnu99', 'gnu11'],
                                                        'none')})
-        if self.compiler_type == CompilerType.GCC_MINGW:
+        if self.compiler_type.is_windows_compiler:
             opts.update({
                 'c_winlibs': coredata.UserArrayOption('c_winlibs', 'Standard Win libraries to link against',
                                                       gnu_winlibs), })
@@ -1150,7 +1214,7 @@ class GnuCCompiler(GnuCompiler, CCompiler):
         return args
 
     def get_option_link_args(self, options):
-        if self.compiler_type == CompilerType.GCC_MINGW:
+        if self.compiler_type.is_windows_compiler:
             return options['c_winlibs'].value[:]
         return []
 
@@ -1175,11 +1239,13 @@ class ElbrusCCompiler(GnuCCompiler, ElbrusCompiler):
 
     # Elbrus C compiler does not have lchmod, but there is only linker warning, not compiler error.
     # So we should explicitly fail at this case.
-    def has_function(self, funcname, prefix, env, extra_args=None, dependencies=None):
+    def has_function(self, funcname, prefix, env, *, extra_args=None, dependencies=None):
         if funcname == 'lchmod':
             return False
         else:
-            return super().has_function(funcname, prefix, env, extra_args, dependencies)
+            return super().has_function(funcname, prefix, env,
+                                        extra_args=extra_args,
+                                        dependencies=dependencies)
 
 
 class IntelCCompiler(IntelCompiler, CCompiler):
@@ -1187,7 +1253,7 @@ class IntelCCompiler(IntelCompiler, CCompiler):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
         IntelCompiler.__init__(self, compiler_type)
         self.lang_header = 'c-header'
-        default_warn_args = ['-Wall', '-w3', '-diag-disable:remark', '-Wpch-messages']
+        default_warn_args = ['-Wall', '-w3', '-diag-disable:remark']
         self.warn_args = {'1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
                           '3': default_warn_args + ['-Wextra']}
@@ -1253,7 +1319,7 @@ class VisualStudioCCompiler(CCompiler):
 
     def get_buildtype_args(self, buildtype):
         args = compilers.msvc_buildtype_args[buildtype]
-        if version_compare(self.version, '<18.0'):
+        if self.id == 'msvc' and version_compare(self.version, '<18.0'):
             args = [arg for arg in args if arg != '/Gw']
         return args
 
@@ -1271,6 +1337,8 @@ class VisualStudioCCompiler(CCompiler):
 
     def get_pch_use_args(self, pch_dir, header):
         base = os.path.basename(header)
+        if self.id == 'clang-cl':
+            base = header
         pchname = self.get_pch_name(header)
         return ['/FI' + base, '/Yu' + base, '/Fp' + os.path.join(pch_dir, pchname)]
 
@@ -1298,7 +1366,12 @@ class VisualStudioCCompiler(CCompiler):
         return []
 
     def get_linker_exelist(self):
-        return ['link'] # FIXME, should have same path as compiler.
+        # FIXME, should have same path as compiler.
+        # FIXME, should be controllable via cross-file.
+        if self.id == 'clang-cl':
+            return ['lld-link']
+        else:
+            return ['link']
 
     def get_linker_always_args(self):
         return ['/nologo']
@@ -1406,6 +1479,8 @@ class VisualStudioCCompiler(CCompiler):
     # http://stackoverflow.com/questions/15259720/how-can-i-make-the-microsoft-c-compiler-treat-unknown-flags-as-errors-rather-t
     def has_arguments(self, args, env, code, mode):
         warning_text = '4044' if mode == 'link' else '9002'
+        if self.id == 'clang-cl' and mode != 'link':
+            args = args + ['-Werror=unknown-argument']
         with self._build_wrapper(code, env, extra_args=args, mode=mode) as p:
             if p.returncode != 0:
                 return False
@@ -1421,7 +1496,7 @@ class VisualStudioCCompiler(CCompiler):
         # build obviously, which is why we only do this when PCH is on.
         # This was added in Visual Studio 2013 (MSVC 18.0). Before that it was
         # always on: https://msdn.microsoft.com/en-us/library/dn502518.aspx
-        if pch and version_compare(self.version, '>=18.0'):
+        if pch and self.id == 'msvc' and version_compare(self.version, '>=18.0'):
             args = ['/FS'] + args
         return args
 
@@ -1438,7 +1513,7 @@ class VisualStudioCCompiler(CCompiler):
     def get_instruction_set_args(self, instruction_set):
         if self.is_64:
             return vs64_instruction_set_args.get(instruction_set, None)
-        if self.version.split('.')[0] == '16' and instruction_set == 'avx':
+        if self.id == 'msvc' and self.version.split('.')[0] == '16' and instruction_set == 'avx':
             # VS documentation says that this exists and should work, but
             # it does not. The headers do not contain AVX intrinsics
             # and the can not be called.
@@ -1446,6 +1521,10 @@ class VisualStudioCCompiler(CCompiler):
         return vs32_instruction_set_args.get(instruction_set, None)
 
     def get_toolset_version(self):
+        if self.id == 'clang-cl':
+            # I have no idea
+            return '14.1'
+
         # See boost/config/compiler/visualc.cpp for up to date mapping
         try:
             version = int(''.join(self.version.split('.')[0:2]))
@@ -1500,11 +1579,24 @@ class VisualStudioCCompiler(CCompiler):
         # false without compiling anything
         return name in ['dllimport', 'dllexport']
 
+    def get_argument_syntax(self):
+        return 'msvc'
+
+    def get_allow_undefined_link_args(self):
+        # link.exe
+        return ['/FORCE:UNRESOLVED']
+
+
+class ClangClCCompiler(VisualStudioCCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrap, is_64):
+        super().__init__(exelist, version, is_cross, exe_wrap, is_64)
+        self.id = 'clang-cl'
+
 
 class ArmCCompiler(ArmCompiler, CCompiler):
-    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        ArmCompiler.__init__(self)
+        ArmCompiler.__init__(self, compiler_type)
 
     def get_options(self):
         opts = CCompiler.get_options(self)
@@ -1519,3 +1611,45 @@ class ArmCCompiler(ArmCompiler, CCompiler):
         if std.value != 'none':
             args.append('--' + std.value)
         return args
+
+class CcrxCCompiler(CcrxCompiler, CCompiler):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, **kwargs):
+        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
+        CcrxCompiler.__init__(self, compiler_type)
+
+    # Override CCompiler.get_always_args
+    def get_always_args(self):
+        return ['-nologo']
+
+    def get_options(self):
+        opts = CCompiler.get_options(self)
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none', 'c89', 'c99'],
+                                                       'none')})
+        return opts
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['c_std']
+        if std.value == 'c89':
+            args.append('-lang=c')
+        elif std.value == 'c99':
+            args.append('-lang=c99')
+        return args
+
+    def get_compile_only_args(self):
+        return []
+
+    def get_no_optimization_args(self):
+        return ['-optimize=0']
+
+    def get_output_args(self, target):
+        return ['-output=obj=%s' % target]
+
+    def get_linker_output_args(self, outputname):
+        return ['-output=%s' % outputname]
+
+    def get_include_args(self, path, is_system):
+        if path == '':
+            path = '.'
+        return ['-include=' + path]
