@@ -154,10 +154,10 @@ class GnomeModule(ExtensionModule):
 
         # Make source dirs relative to build dir now
         source_dirs = [os.path.join(state.build_to_src, state.subdir, d) for d in source_dirs]
-        # Always include current directory, but after paths set by user
-        source_dirs.append(os.path.join(state.build_to_src, state.subdir))
         # Ensure build directories of generated deps are included
         source_dirs += subdirs
+        # Always include current directory, but after paths set by user
+        source_dirs.append(os.path.join(state.build_to_src, state.subdir))
 
         for source_dir in OrderedSet(source_dirs):
             cmd += ['--sourcedir', source_dir]
@@ -327,19 +327,20 @@ class GnomeModule(ExtensionModule):
 
         for dep in deps:
             if isinstance(dep, InternalDependency):
-                cflags.update(dep.compile_args)
+                cflags.update(dep.get_compile_args())
                 cflags.update(get_include_args(dep.include_directories))
                 for lib in dep.libraries:
                     if hasattr(lib, 'held_object'):
                         lib = lib.held_object
-                    internal_ldflags.update(self._get_link_args(state, lib, depends, include_rpath))
-                    libdepflags = self._get_dependencies_flags(lib.get_external_deps(), state, depends, include_rpath,
-                                                               use_gir_args, True)
-                    cflags.update(libdepflags[0])
-                    internal_ldflags.update(libdepflags[1])
-                    external_ldflags.update(libdepflags[2])
-                    external_ldflags_nodedup += libdepflags[3]
-                    gi_includes.update(libdepflags[4])
+                    if isinstance(lib, build.SharedLibrary):
+                        internal_ldflags.update(self._get_link_args(state, lib, depends, include_rpath))
+                        libdepflags = self._get_dependencies_flags(lib.get_external_deps(), state, depends, include_rpath,
+                                                                   use_gir_args, True)
+                        cflags.update(libdepflags[0])
+                        internal_ldflags.update(libdepflags[1])
+                        external_ldflags.update(libdepflags[2])
+                        external_ldflags_nodedup += libdepflags[3]
+                        gi_includes.update(libdepflags[4])
                 extdepflags = self._get_dependencies_flags(dep.ext_deps, state, depends, include_rpath,
                                                            use_gir_args, True)
                 cflags.update(extdepflags[0])
@@ -407,11 +408,21 @@ class GnomeModule(ExtensionModule):
         else:
             return cflags, internal_ldflags, external_ldflags, external_ldflags_nodedup, gi_includes
 
-    def _unwrap_gir_target(self, girtarget):
+    def _unwrap_gir_target(self, girtarget, state):
         while hasattr(girtarget, 'held_object'):
             girtarget = girtarget.held_object
-        if not isinstance(girtarget, (build.Executable, build.SharedLibrary)):
-            raise MesonException('Gir target must be an executable or shared library')
+
+        if not isinstance(girtarget, (build.Executable, build.SharedLibrary,
+                                      build.StaticLibrary)):
+            raise MesonException('Gir target must be an executable or library')
+
+        STATIC_BUILD_REQUIRED_VERSION = ">=1.58.1"
+        if isinstance(girtarget, (build.StaticLibrary)) and \
+           not mesonlib.version_compare(
+               self._get_gir_dep(state)[0].get_version(),
+               STATIC_BUILD_REQUIRED_VERSION):
+            raise MesonException('Static libraries can only be introspected with GObject-Introspection ' + STATIC_BUILD_REQUIRED_VERSION)
+
         return girtarget
 
     def _get_gir_dep(self, state):
@@ -519,7 +530,12 @@ class GnomeModule(ExtensionModule):
         ret = []
 
         for lang in langs:
-            for link_arg in state.environment.coredata.get_external_link_args(lang):
+            if state.environment.is_cross_build():
+                link_args = state.environment.cross_info.config["properties"].get(lang + '_link_args', "")
+            else:
+                link_args = state.environment.coredata.get_external_link_args(lang)
+
+            for link_arg in link_args:
                 if link_arg.startswith('-L'):
                     ret.append(link_arg)
 
@@ -531,8 +547,13 @@ class GnomeModule(ExtensionModule):
         for girtarget in girtargets:
             if isinstance(girtarget, build.Executable):
                 ret += ['--program', girtarget]
-            elif isinstance(girtarget, build.SharedLibrary):
-                libname = girtarget.get_basename()
+            else:
+                # Because of https://gitlab.gnome.org/GNOME/gobject-introspection/merge_requests/72
+                # we can't use the full path until this is merged.
+                if isinstance(girtarget, build.SharedLibrary):
+                    libname = girtarget.get_basename()
+                else:
+                    libname = os.path.join("@PRIVATE_OUTDIR_ABS_%s@" % girtarget.get_id(), girtarget.get_filename())
                 # Needed for the following binutils bug:
                 # https://github.com/mesonbuild/meson/issues/1911
                 # However, g-ir-scanner does not understand -Wl,-rpath
@@ -692,7 +713,10 @@ class GnomeModule(ExtensionModule):
     def _get_external_args_for_langs(self, state, langs):
         ret = []
         for lang in langs:
-            ret += state.environment.coredata.get_external_args(lang)
+            if state.environment.is_cross_build():
+                ret += state.environment.cross_info.config["properties"].get(lang + '_args', "")
+            else:
+                ret += state.environment.coredata.get_external_args(lang)
         return ret
 
     @staticmethod
@@ -723,7 +747,7 @@ class GnomeModule(ExtensionModule):
         giscanner = self.interpreter.find_program_impl('g-ir-scanner')
         gicompiler = self.interpreter.find_program_impl('g-ir-compiler')
 
-        girtargets = [self._unwrap_gir_target(arg) for arg in args]
+        girtargets = [self._unwrap_gir_target(arg, state) for arg in args]
 
         if len(girtargets) > 1 and any([isinstance(el, build.Executable) for el in girtargets]):
             raise MesonException('generate_gir only accepts a single argument when one of the arguments is an executable')
@@ -884,12 +908,14 @@ This will become a hard error in the future.''')
         return ModuleReturnValue(None, rv)
 
     @FeatureNewKwargs('gnome.gtkdoc', '0.48.0', ['c_args'])
+    @FeatureNewKwargs('gnome.gtkdoc', '0.48.0', ['module_version'])
     @FeatureNewKwargs('gnome.gtkdoc', '0.37.0', ['namespace', 'mode'])
     @permittedKwargs({'main_xml', 'main_sgml', 'src_dir', 'dependencies', 'install',
                       'install_dir', 'scan_args', 'scanobjs_args', 'gobject_typesfile',
                       'fixxref_args', 'html_args', 'html_assets', 'content_files',
                       'mkdb_args', 'ignore_headers', 'include_directories',
-                      'namespace', 'mode', 'expand_content_files'})
+                      'namespace', 'mode', 'expand_content_files', 'module_version',
+                      'c_args'})
     def gtkdoc(self, state, args, kwargs):
         if len(args) != 1:
             raise MesonException('Gtkdoc must have one positional argument.')
@@ -904,11 +930,14 @@ This will become a hard error in the future.''')
         main_xml = kwargs.get('main_xml', '')
         if not isinstance(main_xml, str):
             raise MesonException('Main xml keyword argument must be a string.')
+        moduleversion = kwargs.get('module_version', '')
+        if not isinstance(moduleversion, str):
+            raise MesonException('Module version keyword argument must be a string.')
         if main_xml != '':
             if main_file != '':
                 raise MesonException('You can only specify main_xml or main_sgml, not both.')
             main_file = main_xml
-        targetname = modulename + '-doc'
+        targetname = modulename + ('-' + moduleversion if moduleversion else '') + '-doc'
         command = state.environment.get_build_command()
 
         namespace = kwargs.get('namespace', '')
@@ -939,6 +968,7 @@ This will become a hard error in the future.''')
                 '--headerdirs=' + '@@'.join(header_dirs),
                 '--mainfile=' + main_file,
                 '--modulename=' + modulename,
+                '--moduleversion=' + moduleversion,
                 '--mode=' + mode]
         if namespace:
             args.append('--namespace=' + namespace)
@@ -1008,12 +1038,18 @@ This will become a hard error in the future.''')
         ldflags.update(external_ldflags)
 
         if state.environment.is_cross_build():
+            cflags.update(state.environment.cross_info.config["properties"].get('c_args', ""))
+            ldflags.update(state.environment.cross_info.config["properties"].get('c_link_args', ""))
             compiler = state.environment.coredata.cross_compilers.get('c')
         else:
             cflags.update(state.environment.coredata.get_external_args('c'))
             ldflags.update(state.environment.coredata.get_external_link_args('c'))
             compiler = state.environment.coredata.compilers.get('c')
 
+        compiler_flags = self._get_langs_compilers_flags(state, [('c', compiler)])
+        cflags.update(compiler_flags[0])
+        ldflags.update(compiler_flags[1])
+        ldflags.update(compiler_flags[2])
         if compiler:
             args += ['--cc=%s' % ' '.join([shlex.quote(x) for x in compiler.get_exelist()])]
             args += ['--ld=%s' % ' '.join([shlex.quote(x) for x in compiler.get_linker_exelist()])]
