@@ -17,14 +17,15 @@ import json
 
 from pathlib import Path
 from .. import mesonlib
+from ..mesonlib import MesonException
 from . import ExtensionModule
 from mesonbuild.modules import ModuleReturnValue
 from ..interpreterbase import (
     noPosargs, noKwargs, permittedKwargs,
     InterpreterObject, InvalidArguments,
-    FeatureNew
+    FeatureNew, FeatureNewKwargs, disablerIfNotFound
 )
-from ..interpreter import ExternalProgramHolder
+from ..interpreter import ExternalProgramHolder, extract_required_kwarg
 from ..interpreterbase import flatten
 from ..build import known_shmod_kwargs
 from .. import mlog
@@ -64,29 +65,48 @@ class PythonDependency(ExternalDependency):
         else:
             self.major_version = 2
 
+        # We first try to find the necessary python variables using pkgconfig
         if DependencyMethods.PKGCONFIG in self.methods and not python_holder.is_pypy:
             pkg_version = self.variables.get('LDVERSION') or self.version
             pkg_libdir = self.variables.get('LIBPC')
-            old_pkg_libdir = os.environ.get('PKG_CONFIG_LIBDIR')
-            old_pkg_path = os.environ.get('PKG_CONFIG_PATH')
 
-            os.environ.pop('PKG_CONFIG_PATH', None)
+            # If python-X.Y.pc exists in LIBPC, we will try to use it
+            if pkg_libdir is not None and Path(os.path.join(pkg_libdir, 'python-{}.pc'.format(pkg_version))).is_file():
+                old_pkg_libdir = os.environ.get('PKG_CONFIG_LIBDIR')
+                old_pkg_path = os.environ.get('PKG_CONFIG_PATH')
 
-            if pkg_libdir:
-                os.environ['PKG_CONFIG_LIBDIR'] = pkg_libdir
+                os.environ.pop('PKG_CONFIG_PATH', None)
 
-            try:
-                self.pkgdep = PkgConfigDependency('python-{}'.format(pkg_version), environment, kwargs)
-            except Exception:
-                pass
+                if pkg_libdir:
+                    os.environ['PKG_CONFIG_LIBDIR'] = pkg_libdir
 
-            if old_pkg_path is not None:
-                os.environ['PKG_CONFIG_PATH'] = old_pkg_path
+                try:
+                    self.pkgdep = PkgConfigDependency('python-{}'.format(pkg_version), environment, kwargs)
+                    mlog.debug('Found "python-{}" via pkgconfig lookup in LIBPC ({})'.format(pkg_version, pkg_libdir))
+                    py_lookup_method = 'pkgconfig'
+                except MesonException as e:
+                    mlog.debug('"python-{}" could not be found in LIBPC ({})'.format(pkg_version, pkg_libdir))
+                    mlog.debug(e)
 
-            if old_pkg_libdir is not None:
-                os.environ['PKG_CONFIG_LIBDIR'] = old_pkg_libdir
+                if old_pkg_path is not None:
+                    os.environ['PKG_CONFIG_PATH'] = old_pkg_path
+
+                if old_pkg_libdir is not None:
+                    os.environ['PKG_CONFIG_LIBDIR'] = old_pkg_libdir
+                else:
+                    os.environ.pop('PKG_CONFIG_LIBDIR', None)
             else:
-                os.environ.pop('PKG_CONFIG_LIBDIR', None)
+                mlog.debug('"python-{}" could not be found in LIBPC ({}), this is likely due to a relocated python installation'.format(pkg_version, pkg_libdir))
+
+            # If lookup via LIBPC failed, try to use fallback PKG_CONFIG_LIBDIR/PKG_CONFIG_PATH mechanisms
+            if self.pkgdep is None or not self.pkgdep.found():
+                try:
+                    self.pkgdep = PkgConfigDependency('python-{}'.format(pkg_version), environment, kwargs)
+                    mlog.debug('Found "python-{}" via fallback pkgconfig lookup in PKG_CONFIG_LIBDIR/PKG_CONFIG_PATH'.format(pkg_version))
+                    py_lookup_method = 'pkgconfig-fallback'
+                except MesonException as e:
+                    mlog.debug('"python-{}" could not be found via fallback pkgconfig lookup in PKG_CONFIG_LIBDIR/PKG_CONFIG_PATH'.format(pkg_version))
+                    mlog.debug(e)
 
         if self.pkgdep and self.pkgdep.found():
             self.compile_args = self.pkgdep.get_compile_args()
@@ -96,14 +116,19 @@ class PythonDependency(ExternalDependency):
         else:
             self.pkgdep = None
 
+            # Finally, try to find python via SYSCONFIG as a final measure
             if DependencyMethods.SYSCONFIG in self.methods:
                 if mesonlib.is_windows():
                     self._find_libpy_windows(environment)
                 else:
                     self._find_libpy(python_holder, environment)
 
+                if self.is_found:
+                    mlog.debug('Found "python-{}" via SYSCONFIG module'.format(self.version))
+                    py_lookup_method = 'sysconfig'
+
         if self.is_found:
-            mlog.log('Dependency', mlog.bold(self.name), 'found:', mlog.green('YES'))
+            mlog.log('Dependency', mlog.bold(self.name), 'found:', mlog.green('YES ({})'.format(py_lookup_method)))
         else:
             mlog.log('Dependency', mlog.bold(self.name), 'found:', mlog.red('NO'))
 
@@ -126,8 +151,8 @@ class PythonDependency(ExternalDependency):
         largs = self.clib_compiler.find_library(libname, environment, libdirs)
 
         self.is_found = largs is not None
-
-        self.link_args = largs
+        if self.is_found:
+            self.link_args = largs
 
         inc_paths = mesonlib.OrderedSet([
             self.variables.get('INCLUDEPY'),
@@ -235,52 +260,37 @@ class PythonDependency(ExternalDependency):
             return super().get_pkgconfig_variable(variable_name, kwargs)
 
 
-VARIABLES_COMMAND = '''
+INTROSPECT_COMMAND = '''
 import sysconfig
 import json
-
-print (json.dumps (sysconfig.get_config_vars()))
-'''
-
-
-PATHS_COMMAND = '''
-import sysconfig
-import json
-
-print (json.dumps(sysconfig.get_paths()))
-'''
-
-
-INSTALL_PATHS_COMMAND = '''
-import sysconfig
-import json
-
-print (json.dumps(sysconfig.get_paths(scheme='posix_prefix', vars={'base': '', 'platbase': '', 'installed_base': ''})))
-'''
-
-
-IS_PYPY_COMMAND = '''
 import sys
-import json
 
-print (json.dumps('__pypy__' in sys.builtin_module_names))
+install_paths = sysconfig.get_paths(scheme='posix_prefix', vars={'base': '', 'platbase': '', 'installed_base': ''})
+
+print (json.dumps ({
+  'variables': sysconfig.get_config_vars(),
+  'paths': sysconfig.get_paths(),
+  'install_paths': install_paths,
+  'version': sysconfig.get_python_version(),
+  'platform': sysconfig.get_platform(),
+  'is_pypy': '__pypy__' in sys.builtin_module_names,
+}))
 '''
 
-
 class PythonInstallation(ExternalProgramHolder, InterpreterObject):
-    def __init__(self, interpreter, python):
+    def __init__(self, interpreter, python, info):
         InterpreterObject.__init__(self)
         ExternalProgramHolder.__init__(self, python)
         self.interpreter = interpreter
         prefix = self.interpreter.environment.coredata.get_builtin_option('prefix')
-        self.variables = json.loads(run_command(python, VARIABLES_COMMAND))
-        self.paths = json.loads(run_command(python, PATHS_COMMAND))
-        install_paths = json.loads(run_command(python, INSTALL_PATHS_COMMAND))
+        self.variables = info['variables']
+        self.paths = info['paths']
+        install_paths = info['install_paths']
         self.platlib_install_path = os.path.join(prefix, install_paths['platlib'][1:])
         self.purelib_install_path = os.path.join(prefix, install_paths['purelib'][1:])
-        self.version = run_command(python, "import sysconfig; print (sysconfig.get_python_version())")
-        self.platform = run_command(python, "import sysconfig; print (sysconfig.get_platform())")
-        self.is_pypy = json.loads(run_command(python, IS_PYPY_COMMAND))
+        self.version = info['version']
+        self.platform = info['platform']
+        self.is_pypy = info['is_pypy']
 
     @permittedKwargs(mod_kwargs)
     def extension_module(self, interpreter, state, args, kwargs):
@@ -450,16 +460,29 @@ class PythonModule(ExtensionModule):
         else:
             return None
 
+    def _check_version(self, name_or_path, version):
+        if name_or_path == 'python2':
+            return mesonlib.version_compare(version, '< 3.0')
+        elif name_or_path == 'python3':
+            return mesonlib.version_compare(version, '>= 3.0')
+        return True
+
+    @FeatureNewKwargs('python.find_installation', '0.49.0', ['disabler'])
+    @disablerIfNotFound
     @permittedKwargs(['required'])
     def find_installation(self, interpreter, state, args, kwargs):
-        required = kwargs.get('required', True)
-        if not isinstance(required, bool):
-            raise InvalidArguments('"required" argument must be a boolean.')
+        feature_check = FeatureNew('Passing "feature" option to find_installation', '0.48.0')
+        disabled, required, feature = extract_required_kwarg(kwargs, state.subproject, feature_check)
+        if disabled:
+            mlog.log('find_installation skipped: feature', mlog.bold(feature), 'disabled')
+            return ExternalProgramHolder(NonExistingExternalProgram())
 
         if len(args) > 1:
             raise InvalidArguments('find_installation takes zero or one positional argument.')
 
-        if args:
+        if 'python' in state.environment.config_info.binaries:
+            name_or_path = state.environment.config_info.binaries['python']
+        elif args:
             name_or_path = args[0]
             if not isinstance(name_or_path, str):
                 raise InvalidArguments('find_installation argument must be a string.')
@@ -470,23 +493,20 @@ class PythonModule(ExtensionModule):
             mlog.log("Using meson's python {}".format(mesonlib.python_command))
             python = ExternalProgram('python3', mesonlib.python_command, silent=True)
         else:
-            if mesonlib.is_windows():
+            python = ExternalProgram(name_or_path, silent = True)
+
+            if not python.found() and mesonlib.is_windows():
                 pythonpath = self._get_win_pythonpath(name_or_path)
                 if pythonpath is not None:
                     name_or_path = pythonpath
-            python = ExternalProgram(name_or_path, silent = True)
+                    python = ExternalProgram(name_or_path, silent = True)
+
             # Last ditch effort, python2 or python3 can be named python
             # on various platforms, let's not give up just yet, if an executable
             # named python is available and has a compatible version, let's use
             # it
             if not python.found() and name_or_path in ['python2', 'python3']:
                 python = ExternalProgram('python', silent = True)
-                if python.found():
-                    version = run_command(python, "import sysconfig; print (sysconfig.get_python_version())")
-                    if not version or \
-                            name_or_path == 'python2' and mesonlib.version_compare(version, '>= 3.0') or \
-                            name_or_path == 'python3' and not mesonlib.version_compare(version, '>= 3.0'):
-                        python = NonExistingExternalProgram()
 
         if not python.found():
             if required:
@@ -494,13 +514,17 @@ class PythonModule(ExtensionModule):
             res = ExternalProgramHolder(NonExistingExternalProgram())
         else:
             # Sanity check, we expect to have something that at least quacks in tune
-            version = run_command(python, "import sysconfig; print (sysconfig.get_python_version())")
-            if not version:
+            try:
+                info = json.loads(run_command(python, INTROSPECT_COMMAND))
+            except json.JSONDecodeError:
+                info = None
+
+            if isinstance(info, dict) and 'version' in info and self._check_version(name_or_path, info['version']):
+                res = PythonInstallation(interpreter, python, info)
+            else:
                 res = ExternalProgramHolder(NonExistingExternalProgram())
                 if required:
                     raise mesonlib.MesonException('{} is not a valid python'.format(python))
-            else:
-                res = PythonInstallation(interpreter, python)
 
         return res
 
