@@ -20,9 +20,10 @@ import pickle
 from mesonbuild import build
 from mesonbuild import environment
 from mesonbuild.dependencies import ExternalProgram
-from mesonbuild.mesonlib import MesonException
+from mesonbuild.mesonlib import substring_is_in_list, MesonException
 from mesonbuild import mlog
 
+import tempfile
 import time, datetime, multiprocessing, json
 import concurrent.futures as conc
 import platform
@@ -60,8 +61,7 @@ def determine_worker_count():
             num_workers = 1
     return num_workers
 
-def buildparser():
-    parser = argparse.ArgumentParser(prog='meson test')
+def add_arguments(parser):
     parser.add_argument('--repeat', default=1, dest='repeat', type=int,
                         help='Number of times to run the tests.')
     parser.add_argument('--no-rebuild', default=False, action='store_true',
@@ -102,7 +102,6 @@ def buildparser():
                         help='Arguments to pass to the specified test(s) or all tests')
     parser.add_argument('args', nargs='*',
                         help='Optional list of tests to run')
-    return parser
 
 
 def returncode_to_status(retcode):
@@ -145,6 +144,8 @@ class TestResult(enum.Enum):
     TIMEOUT = 'TIMEOUT'
     SKIP = 'SKIP'
     FAIL = 'FAIL'
+    EXPECTEDFAIL = 'EXPECTEDFAIL'
+    UNEXPECTEDPASS = 'UNEXPECTEDPASS'
 
 
 class TestRun:
@@ -244,7 +245,11 @@ class SingleTestRunner:
                     # because there is no execute wrapper.
                     return None
                 else:
-                    return [self.test.exe_runner] + self.test.fname
+                    if not self.test.exe_runner.found():
+                        msg = 'The exe_wrapper defined in the cross file {!r} was not ' \
+                              'found. Please check the command and/or add it to PATH.'
+                        raise TestException(msg.format(self.test.exe_runner.name))
+                    return self.test.exe_runner.get_command() + self.test.fname
             else:
                 return self.test.fname
 
@@ -261,19 +266,12 @@ class SingleTestRunner:
                 self.test.timeout = None
             return self._run_cmd(wrap + cmd + self.test.cmd_args + self.options.test_args)
 
-    @staticmethod
-    def _substring_in_list(substr, strlist):
-        for s in strlist:
-            if substr in s:
-                return True
-        return False
-
     def _run_cmd(self, cmd):
         starttime = time.time()
 
         if len(self.test.extra_paths) > 0:
             self.env['PATH'] = os.pathsep.join(self.test.extra_paths + ['']) + self.env['PATH']
-            if self._substring_in_list('wine', cmd):
+            if substring_is_in_list('wine', cmd):
                 wine_paths = ['Z:' + p for p in self.test.extra_paths]
                 wine_path = ';'.join(wine_paths)
                 # Don't accidentally end with an `;` because that will add the
@@ -294,8 +292,8 @@ class SingleTestRunner:
         stdout = None
         stderr = None
         if not self.options.verbose:
-            stdout = subprocess.PIPE
-            stderr = subprocess.PIPE if self.options and self.options.split else subprocess.STDOUT
+            stdout = tempfile.TemporaryFile("wb+")
+            stderr = tempfile.TemporaryFile("wb+") if self.options and self.options.split else stdout
 
         # Let gdb handle ^C instead of us
         if self.options.gdb:
@@ -329,7 +327,7 @@ class SingleTestRunner:
         else:
             timeout = self.test.timeout
         try:
-            (stdo, stde) = p.communicate(timeout=timeout)
+            p.communicate(timeout=timeout)
         except subprocess.TimeoutExpired:
             if self.options.verbose:
                 print('%s time out (After %d seconds)' % (self.test.name, timeout))
@@ -342,6 +340,8 @@ class SingleTestRunner:
                 # Let us accept ^C again
                 signal.signal(signal.SIGINT, previous_sigint_handler)
 
+        additional_error = None
+
         if kill_test or timed_out:
             # Python does not provide multiplatform support for
             # killing a process and all its children so we need
@@ -357,20 +357,44 @@ class SingleTestRunner:
                     # There's nothing we can do (maybe the process
                     # already died) so carry on.
                     pass
-            (stdo, stde) = p.communicate()
+            try:
+                p.communicate(timeout=1)
+            except subprocess.TimeoutExpired:
+                # An earlier kill attempt has not worked for whatever reason.
+                # Try to kill it one last time with a direct call.
+                # If the process has spawned children, they will remain around.
+                p.kill()
+                try:
+                    p.communicate(timeout=1)
+                except subprocess.TimeoutExpired:
+                    additional_error = b'Test process could not be killed.'
+            except ValueError:
+                additional_error = b'Could not read output. Maybe the process has redirected its stdout/stderr?'
         endtime = time.time()
         duration = endtime - starttime
-        stdo = decode(stdo)
-        if stde:
-            stde = decode(stde)
+        if additional_error is None:
+            if stdout is None:  # if stdout is None stderr should be as well
+                stdo = ''
+                stde = ''
+            else:
+                stdout.seek(0)
+                stdo = decode(stdout.read())
+                if stderr != stdout:
+                    stderr.seek(0)
+                    stde = decode(stderr.read())
+                else:
+                    stde = ""
+        else:
+            stdo = ""
+            stde = additional_error
         if timed_out:
             res = TestResult.TIMEOUT
         elif p.returncode == GNU_SKIP_RETURNCODE:
             res = TestResult.SKIP
-        elif self.test.should_fail == bool(p.returncode):
-            res = TestResult.OK
+        elif self.test.should_fail:
+            res = TestResult.EXPECTEDFAIL if bool(p.returncode) else TestResult.UNEXPECTEDPASS
         else:
-            res = TestResult.FAIL
+            res = TestResult.FAIL if bool(p.returncode) else TestResult.OK
         return TestRun(res, p.returncode, self.test.should_fail, duration, stdo, stde, cmd, self.test.env)
 
 
@@ -379,6 +403,8 @@ class TestHarness:
         self.options = options
         self.collected_logs = []
         self.fail_count = 0
+        self.expectedfail_count = 0
+        self.unexpectedpass_count = 0
         self.success_count = 0
         self.skip_count = 0
         self.timeout_count = 0
@@ -424,6 +450,8 @@ class TestHarness:
 
     def get_test_runner(self, test):
         options = deepcopy(self.options)
+        if not options.setup:
+            options.setup = self.build_data.test_setup_default_name
         if options.setup:
             env = self.merge_suite_options(options, test)
         else:
@@ -443,6 +471,10 @@ class TestHarness:
             self.success_count += 1
         elif result.res is TestResult.FAIL:
             self.fail_count += 1
+        elif result.res is TestResult.EXPECTEDFAIL:
+            self.expectedfail_count += 1
+        elif result.res is TestResult.UNEXPECTEDPASS:
+            self.unexpectedpass_count += 1
         else:
             sys.exit('Unknown test result encountered: {}'.format(result.res))
 
@@ -458,9 +490,10 @@ class TestHarness:
         result_str = '%s %s  %s%s%s%5.2f s %s' % \
             (num, name, padding1, result.res.value, padding2, result.duration,
              status)
-        if not self.options.quiet or result.res is not TestResult.OK:
-            if result.res is not TestResult.OK and mlog.colorize_console:
-                if result.res in (TestResult.FAIL, TestResult.TIMEOUT):
+        ok_statuses = (TestResult.OK, TestResult.EXPECTEDFAIL)
+        if not self.options.quiet or result.res not in ok_statuses:
+            if result.res not in ok_statuses and mlog.colorize_console:
+                if result.res in (TestResult.FAIL, TestResult.TIMEOUT, TestResult.UNEXPECTEDPASS):
                     decorator = mlog.red
                 elif result.res is TestResult.SKIP:
                     decorator = mlog.yellow
@@ -481,11 +514,14 @@ class TestHarness:
 
     def print_summary(self):
         msg = '''
-OK:      %4d
-FAIL:    %4d
-SKIP:    %4d
-TIMEOUT: %4d
-''' % (self.success_count, self.fail_count, self.skip_count, self.timeout_count)
+Ok:                 %4d
+Expected Fail:      %4d
+Fail:               %4d
+Unexpected Pass:    %4d
+Skipped:            %4d
+Timeout:            %4d
+''' % (self.success_count, self.expectedfail_count, self.fail_count,
+            self.unexpectedpass_count, self.skip_count, self.timeout_count)
         print(msg)
         if self.logfile:
             self.logfile.write(msg)
@@ -503,7 +539,11 @@ TIMEOUT: %4d
                     print('--- Listing only the last 100 lines from a long log. ---')
                     lines = lines[-100:]
                 for line in lines:
-                    print(line)
+                    try:
+                        print(line)
+                    except UnicodeEncodeError:
+                        line = line.encode('ascii', errors='replace').decode()
+                        print(line)
 
     def doit(self):
         if self.is_run:
@@ -726,9 +766,7 @@ def rebuild_all(wd):
 
     return True
 
-def run(args):
-    options = buildparser().parse_args(args)
-
+def run(options):
     if options.benchmark:
         options.num_processes = 1
 
@@ -750,12 +788,13 @@ def run(args):
     if check_bin is not None:
         exe = ExternalProgram(check_bin, silent=True)
         if not exe.found():
-            sys.exit('Could not find requested program: %s' % check_bin)
+            print('Could not find requested program: {!r}'.format(check_bin))
+            return 1
     options.wd = os.path.abspath(options.wd)
 
     if not options.list and not options.no_rebuild:
         if not rebuild_all(options.wd):
-            sys.exit(-1)
+            return 1
 
     try:
         th = TestHarness(options)
@@ -767,5 +806,14 @@ def run(args):
         return th.run_special()
     except TestException as e:
         print('Meson test encountered an error:\n')
-        print(e)
+        if os.environ.get('MESON_FORCE_BACKTRACE'):
+            raise e
+        else:
+            print(e)
         return 1
+
+def run_with_args(args):
+    parser = argparse.ArgumentParser(prog='meson test')
+    add_arguments(parser)
+    options = parser.parse_args(args)
+    return run(options)
